from __future__ import annotations

import importlib.resources
import itertools
import subprocess
import sys
from pathlib import Path
from tempfile import TemporaryDirectory
from typing import Any

from pdm.cli.actions import resolve_candidates_from_lockfile
from pdm.environments import PythonEnvironment
from pdm.project import Project


def get_in_process_script():
    if sys.version_info >= (3, 9):
        script = importlib.resources.files("pdm_packer") / "_compile_source.py"
        return importlib.resources.as_file(script)
    else:
        return importlib.resources.path("pdm_packer", "_compile_source.py")


IN_PROCESS_SCRIPT = Path(__file__).with_name("_compile_source.py")
PDM_VERSION = tuple(
    map(
        int,
        itertools.takewhile(str.isdigit, importlib.metadata.version("pdm").split(".")),
    )
)


class PackEnvironment(PythonEnvironment):
    def __init__(self, project: Project) -> None:
        self._dir = TemporaryDirectory(prefix="pdm-pack-")
        super().__init__(project, prefix=self._dir.name)

    def __enter__(self) -> PackEnvironment:
        return self

    def __exit__(self, *args: Any) -> None:
        self._dir.cleanup()

    def _compile_to_pyc(self, dest: Path) -> None:
        with get_in_process_script() as scriptpath:
            args = [str(self.interpreter.path), str(scriptpath), str(dest)]
            subprocess.check_output(args, stderr=subprocess.STDOUT)

    def prepare_lib_for_pack(self, compile: bool = False) -> Path:
        """Get a lib path containing all dependencies for pack.
        Editable packages will be replaced by non-editable ones.
        """
        project = self.project
        this_paths = self.get_paths()
        requirements = project.get_dependencies().values()
        if PDM_VERSION >= (2, 19):
            from pdm.cli.actions import resolve_from_lockfile

            packages = resolve_from_lockfile(project, requirements, groups=["default"])
            synchronizer = project.core.get_synchronizer()(
                self,
                install_self=bool(project.name),
                no_editable=True,
                packages=packages,
            )
        else:
            candidates = resolve_candidates_from_lockfile(
                project, requirements, groups=["default"]
            )
            synchronizer = project.core.synchronizer_class(
                candidates, self, install_self=bool(project.name), no_editable=True
            )
        synchronizer.synchronize()
        dest = Path(this_paths["purelib"])
        if compile:
            self._compile_to_pyc(dest)
        return dest



"""Configuration for the pytest test suite."""

import pytest
from click.testing import CliRunner
from pdm.core import Core


@pytest.fixture(scope="session")
def main() -> Core:
    return Core()


@pytest.fixture(scope="session")
def invoke(main: Core):
    runner = CliRunner(mix_stderr=False)

    def caller(args, *, raising=True, **extras):
        result = runner.invoke(
            main, args, prog_name="pdm", catch_exceptions=not raising, **extras
        )
        if result.exit_code != 0 and raising:
            raise RuntimeError(
                f"Calling command {args} failed with exit code: {result.exit_code}\n"
                f"{result.stderr}"
            )
        return result

    return caller



import argparse
import importlib.resources
import io
import os
import stat
import zipapp
from pathlib import Path

from pdm.cli.commands.base import BaseCommand
from pdm.exceptions import PdmUsageError
from pdm.project import Project

from .env import PackEnvironment


class PackCommand(BaseCommand):
    """Pack the packages into a zipapp"""

    name = "pack"

    def add_arguments(self, parser: argparse.ArgumentParser) -> None:
        parser.add_argument(
            "-m", "--main", help="Specify the console script entry point for the zipapp"
        )
        parser.add_argument(
            "-o",
            "--output",
            help="Specify the output filename, default: the project name",
            type=Path,
        )
        parser.add_argument(
            "-c",
            "--compress",
            action="store_true",
            help="Compress files with the deflate method, no compress by default",
        )
        parser.add_argument(
            "--pyc",
            "--compile",
            dest="compile",
            action="store_true",
            help="Compile source into pyc files",
        )
        parser.add_argument(
            "--no-py",
            action="store_true",
            help="Remove the .py files in favor of .pyc files",
        )
        parser.add_argument(
            "-i",
            "--interpreter",
            help="The Python interpreter path, default: the project interpreter",
        )
        parser.add_argument(
            "--exe",
            action="store_true",
            help="Create an executable file. If the output file isn't given, "
            "the file name will end with .exe(Windows) or no suffix(Posix)",
        )

    @staticmethod
    def _write_zipapp(
        stream: io.BytesIO, project: Project, options: argparse.Namespace
    ) -> Path:
        bytes = stream.getvalue()
        if options.exe and (
            os.name == "nt" or (os.name == "java" and os._name == "nt")
        ):
            interpreter = options.interpreter or project.python.executable
            bits = "32" if project.python.is_32bit else "64"
            kind = "w" if "pythonw" in Path(interpreter).name else "t"
            launcher = importlib.resources.read_binary("distlib", f"{kind}{bits}.exe")
            bytes = launcher + bytes

        if options.output:
            output = options.output
        else:
            name = project.name or project.root.name
            name = name.replace("-", "_")
            suffix = ".pyz" if not options.exe else ".exe" if os.name == "nt" else ""
            output = Path(name + suffix)

        output.write_bytes(bytes)
        if options.exe:
            output.chmod(output.stat().st_mode | stat.S_IEXEC)
        return output

    def handle(self, project: Project, options: argparse.Namespace) -> None:
        def file_filter(name: str) -> bool:
            path = Path(name)
            last = path.name
            return not (
                "__pycache__" in path.parts
                or options.no_py
                and last.endswith(".py")
                or not options.compile
                and last.endswith(".pyc")
            )

        main = None
        if options.no_py and not options.compile:
            raise PdmUsageError("--no-py must be used with --pyc/--compile")
        if options.main:
            main = options.main
        else:
            scripts = project.pyproject.metadata.get("scripts", {})
            if scripts:
                main = str(scripts[next(iter(scripts))])

        target_stream = io.BytesIO()

        with PackEnvironment(project) as pack_env:
            project.core.ui.echo("Packing packages...")
            lib = pack_env.prepare_lib_for_pack(compile=options.compile)
            project.core.ui.echo(f"Packages are prepared at {lib}")
            project.core.ui.echo("Creating zipapp...")
            zipapp.create_archive(
                lib,
                target_stream,
                interpreter=options.interpreter or str(project.python.executable),
                main=main,
                compressed=options.compress,
                filter=file_filter,
            )
            output = self._write_zipapp(target_stream, project, options)
            project.core.ui.echo(f"Zipapp is generated at [green]{output}[/]")



import os
import shutil
import stat
import subprocess
import sys
import zipfile
from pathlib import Path

import pytest
from pdm.utils import cd, is_editable

from pdm_packer.env import PackEnvironment


@pytest.fixture(scope="module")
def example_project(invoke, main):
    tmp_path = Path(__file__).parent / ".testing"
    if tmp_path.exists():
        shutil.rmtree(tmp_path)
    tmp_path.mkdir()
    tmp_path.joinpath("app.py").write_text(
        "import requests\ndef main():\n    print(requests.__version__)\n"
    )
    project = main.create_project(tmp_path)
    project.pyproject.set_data(
        {
            "project": {
                "name": "test_app",
                "version": "0.1.0",
                "requires-python": ">=3.7",
                "dependencies": ["requests==2.31.0"],
            },
            "build-system": {
                "requires": ["pdm-backend"],
                "build-backend": "pdm.backend",
            },
        }
    )
    project.pyproject.write()
    invoke(["lock"], obj=project)

    return project


def test_pack_env_all_non_editable(example_project):
    with PackEnvironment(example_project) as env:
        env.prepare_lib_for_pack()
        for _, v in env.get_working_set().items():
            assert not is_editable(v)


def test_create_without_main_error(example_project, invoke):
    result = invoke(["pack"], raising=False, obj=example_project)
    assert result.exit_code != 0


def test_no_py_not_with_compile(example_project, invoke):
    result = invoke(["pack", "--no-py"], raising=False, obj=example_project)
    assert result.exit_code != 0


def test_create_normal_pyz(example_project, invoke, tmp_path):
    with cd(tmp_path):
        invoke(["pack", "-m", "app:main"], obj=example_project)
    output = tmp_path / "test_app.pyz"
    assert output.exists()
    assert f"#!{example_project.python.executable}".encode() in output.read_bytes()

    with zipfile.ZipFile(output) as zf:
        namelist = zf.namelist()
        assert "requests/__init__.py" in namelist
        assert "urllib3/__init__.py" in namelist
        assert "requests-2.31.0.dist-info/METADATA" in namelist
        assert "app.py" in namelist
        assert not any(name.endswith(".pyc") for name in namelist)

        main = [
            line.decode().rstrip()
            for line in zf.open("__main__.py")
            if not line.startswith(b"#")
        ]
        assert main == ["import app", "app.main()"]

    subprocess.check_call([sys.executable, str(output)])


def test_create_pyz_with_pyc(example_project, invoke, tmp_path):
    with cd(tmp_path):
        invoke(["pack", "-v", "-m", "app:main", "--compile"], obj=example_project)
    output = tmp_path / "test_app.pyz"
    assert output.exists()

    with zipfile.ZipFile(output) as zf:
        namelist = zf.namelist()
        assert "app.pyc" in namelist
        assert "requests/__init__.pyc" in namelist
        assert "requests/__init__.py" in namelist
        assert "urllib3/__init__.pyc" in namelist


def test_create_pyz_without_py(example_project, invoke, tmp_path):
    with cd(tmp_path):
        invoke(
            ["pack", "-v", "-m", "app:main", "--compile", "--no-py"],
            obj=example_project,
        )
    output = tmp_path / "test_app.pyz"
    assert output.exists()

    with zipfile.ZipFile(output) as zf:
        namelist = zf.namelist()
        assert "app.pyc" in namelist
        assert "requests/__init__.pyc" in namelist
        assert "requests/__init__.py" not in namelist
        assert "urllib3/__init__.pyc" in namelist


def test_pack_respect_console_script(example_project, invoke, tmp_path):
    example_project.pyproject.metadata["scripts"] = {"app": "app:main"}
    example_project.pyproject.write()
    with cd(tmp_path):
        invoke(["pack"], obj=example_project)
    output = tmp_path / "test_app.pyz"
    assert output.exists()

    with zipfile.ZipFile(output) as zf:
        main = [
            line.decode().rstrip()
            for line in zf.open("__main__.py")
            if not line.startswith(b"#")
        ]
        assert main == ["import app", "app.main()"]

    subprocess.check_call([sys.executable, str(output)])


def test_pack_change_output_file(example_project, invoke, tmp_path):
    output = tmp_path / "foo.pyz"
    invoke(["pack", "-m", "app:main", "-o", str(output)], obj=example_project)

    assert output.exists()
    subprocess.check_call([sys.executable, str(output)])


def test_pack_create_exe_file(example_project, invoke, tmp_path):
    with cd(tmp_path):
        invoke(["pack", "-m", "app:main", "--exe"], obj=example_project)
    output = tmp_path / ("test_app.exe" if os.name == "nt" else "test_app")
    assert output.exists()
    assert output.stat().st_mode & stat.S_IEXEC

    subprocess.check_call([str(output)])



import os
import py_compile
import sys

IGNORE_DIRS = ["__pycache__"]


def main(dest):
    # type: (str) -> None
    for root, _, files in os.walk(dest):
        if os.path.basename(root) in IGNORE_DIRS:
            continue
        for filename in files:
            if not filename.endswith(".py"):
                continue
            src_file = os.path.join(root, filename)
            dest_file = src_file + "c"
            py_compile.compile(src_file, dest_file)


if __name__ == "__main__":
    main(sys.argv[1])



"""
    pdm-packer

    A PDM plugin that packs your packages into a zipapp
    :author: Frost Ming <mianghong@gmail.com>
    :license: MIT
"""

from pdm.core import Core

from .command import PackCommand


def plugin(core: Core) -> None:
    core.register_command(PackCommand)
